#!/usr/bin/perl
#
# Convert *.bdf to iso10646 encoding and combine missing glyphs
#
# Usage:
# $0 -a ja ./jiskan16.bdf ./hanglg16.bdf > 16x16utf.bdf
#
# -b, -bl	bolden, overstriking to right
# -br		bolden, overstriking to left
# +b		bolden off
#
# Caveats:
# -b (bolden) would not work for >32 pixel fonts
#
use Encode;

$bolden = 0;
$add_style = "";

$usage = "Usage: perl this.pl ./jiskan16.bdf ./hanglg16.bdf > 16x16utf.bdf";
$charsdir = "./uchars";	#tmpdir XXX not parallel safe

if ($ARGV[0] eq "-a") {shift; $add_style = shift;}
if ($ARGV[0] eq "-b") {$bolden = 1; shift;}
if ($ARGV[0] eq "-br") {$bolden = 1; shift;}
if ($ARGV[0] eq "-bl") {$bolden = -1; shift;}
if ($ARGV[0] eq "+b") {$bolden = 0; shift;}
if ($ARGV[0] eq "-1") {$add_scanline = 1; shift;}

-d $charsdir || mkdir($charsdir) || die "cannot create tmpdir $charsdir";

$copyright = "";

## $utf16codepoint = charset2utf16c($ku * 256 + $ten, "JISX0208.1983-0");
## return "0" if conversion fails
sub charset2utf16c 
{
	local ($c,$_) = (@_);
	# translate $c to UTF16 codepoint
	local $utf16be;
	if (/^JISX0208/i || /^JISX0213\.[\d]+-1/i) {
		$utf16be = Encode::encode("utf16be", Encode::decode("iso-2022-jp", 
		sprintf("\033\$B%c%c", ($c >> 8)&255, $c & 255)
		));
	} elsif (/^KSC5601\./i) {
		$utf16be = Encode::encode("utf16be", Encode::decode("iso-2022-kr", 
		sprintf("\033\$)C\x0e%c%c\x0f", ($c >> 8)&255, $c & 255)
		));
	} elsif (/^GB2312/i) {
		$utf16be = Encode::encode("utf16be", Encode::decode("gb2312-raw", 
		sprintf("%c%c", ($c >> 8)&255, $c & 255)
		));
	} elsif (/^ISO8859-1/i) {
		$utf16be = Encode::encode("utf16be", Encode::decode("iso8859-1", 
		sprintf("%c", $c & 255)
		));
	} elsif (/^JISX0201\.1976/i) {
		$utf16be = Encode::encode("utf16be", Encode::decode("jis0201-raw", 
		sprintf("%c", $c & 255)
		));
	} elsif (/^ISO10646/i) {
		$utf16be = sprintf("%c%c", ($c >> 8)&255, $c & 255);
		#XXX overflow 16bit input
	} else {
		die "unrecognized CHARSET_REGISTRY $_";
	}
	@_ = split('',$utf16be);
	if (($#_+1) != 2) {
		print STDERR "FAILED to decode ENCODING $c (0x".sprintf("%04x",$c)."), returned $utf16be (length ".($#_+1).")\n";
		return 0;
	}
	local $c16 = ord($_[0])*256 + ord($_[1]);

	#print STDERR "UTF16BE: ".join(":",map {sprintf("%02x",ord)} split('',$utf16be))."\n";
	#print STDERR "ENCODING $c (0x".sprintf("%04x",$c).") (len ".($#_+1).") translated to $c16\n";

	return $c16;
}

# $_ = embold($bitmap_line, $shift_direction, $bbw)
# $shift_direction= 1 (right), -1 (left)
sub embold {
	local ($_,$dir, $bbw) = @_;
	if ($dir == 0) { return $_; } # do nothing
	if (!/^([0-9A-F]+)\n*$/i) {
		die "Non-BITMAP line \"$_\" passed to embold";
	}

	chomp;
	local $digits=length($_);
	local $right_pad_bits = length($1)*4 - $bbw;
	local $o = hex($_);
	if ($dir == 1) {
		# overstrike to right, detect pile on left
		$o = ~($o>>2) & $o | ($o >> 1);
	} elsif ($dir == -1) {
		# overstrike to left, detect pile on left
		$o = ~($o>>1) & ($o<<1) | $o;
	}
	# limit to BBw bits; MSB packed!
	$o &= ((1<<$bbw)-1) << $right_pad_bits;
	$_ = sprintf("%0*X\n", $digits, $o);
	return $_;
}

# first, open jiskan16.bdf and divide glyphs out into ./uchars/NNNNN
$bdf = shift(@ARGV) || die $usage;
open(BDF, "<", $bdf);
$state = "header";
open(META, ">", "$charsdir/+HEADER") || die "cannot write +HEADER";
print STDERR "Reading $bdf...\n";
while(<BDF>) {
	if (/^FONTBOUNDINGBOX (\d+) (\d+)/) {
		$font_bbw=$1; $font_bbh=$2; #used for sanity check
	}
	if (/^COPYRIGHT "(.*)"/) {
		if ($copyright ne "") { $copyright .= "; "; }
		$copyright .= $1;
	}
	if (/^CHARSET_REGISTRY "(.*)"/) {
		$charset_registry = $1;
	}
	if (/^CHARSET_ENCODING "(.*)"/) {
		$charset_encoding = $1;
	}
	if (/^DEFAULT_CHAR (\d+)/) {
		$default_char = $1;
	}

	if (/^STARTCHAR/) {
		undef $code;
		$hcount=0;
		$cspec = $_;
		# read until ENDCHAR, and dump to separate file
		STARTCHAR: while(<BDF>) {
			if (/^ENCODING ([0-9]+)/) { $code = $1; }
			if (/^BBX ([0-9]+)/) {
				$bbw=$1;
				if ($bbw != $font_bbw) {
					print STDERR "WARNING: BBW $bbw != global $font_bbw in ENCODING $code\n";
				}
			}
			if (/^BITMAP/) {
				$cspec .= $_;
				while(<BDF>) {
					if (/^ENDCHAR/) {
						$cspec .= $_;
						last STARTCHAR;
					}
					$_ = embold($_, $bolden, $bbw);
					$cspec .= $_;
					$hcount++;
				}
			} #BITMAP
			last STARTCHAR if (/^ENDCHAR/);
			$cspec .= $_;
		} #STARTCHAR
		if (! defined($code)) { die "undefined ENCODING in STARTCHAR above line ".$.; }
		if ($hcount != $font_bbh) {
			print STDERR "WARNING: Height $hcount != global $font_bbh in ENCODING $code\n";
		}

		$ucode = charset2utf16c($code, "$charset_registry-$charset_encoding");
		if ($ucode != 0 || $code == 0) {
			$cspec =~ s/^ENCODING .*/"ENCODING $ucode"/mei;
			# dump to a separate file
			open(CSPEC,">", sprintf("$charsdir/%05s",$ucode));
			print CSPEC $cspec;
			close(CSPEC);
		}

		# state transition
		if ($state eq "header") {
			close(META);
			open(META, ">", "$charsdir/=TRAILER") || die "cannot write =TRAILER";
			$state = "trailer";
		}

		next;
	}#STARTCHAR

	print META $_;
}
close BDF;
close META;

## translate $default_char
$default_char = charset2utf16c($default_char, "$charset_registry-$charset_encoding");

## copyglyph($utf16codepoint_from, $utf16codepoint_to)
sub copyglyph 
{
	local ($from,$to) = @_;
	open(COPYF, "<", sprintf("%s/%05s", $charsdir, $from)) || die "copyglyph: glyph $from not found";
	local $/;
	local $_ = <COPYF>;
	close COPYF;

	s/^ENCODING .*/"ENCODING ".(0+$to)/mei;

	# unconditional overwrite (TODO: not overwrite)
	open(COPYT, ">", sprintf("%s/%05s", $charsdir, $to));
	print COPYT $_;
	close COPYT;

	#print STDERR "copyglyph $from to $to\n";
}

if ($charset_registry =~ /^JISX0208\./i) {
	#copyglyph(0x2032, 0x27);	#single quote
	#copyglyph(0x2019, 0x27);	#single quote
	copyglyph(0x2019, 0xff07);	#fullwidth single quote
	#copyglyph(0x2033, 0x22);	#double quote
	#copyglyph(0x201d, 0x22);	#double quote
	copyglyph(0x201d, 0xff02);	#fullwidth double quote

	copyglyph(0x2212, 0xff0d);	#fullwidth minus
	copyglyph(0x301c, 0xff5e);	#fullwidth tilde
	copyglyph(0xa2, 0xffe0);	#fullwidth cent
	copyglyph(0xa3, 0xffe1);	#fullwidth pound
	copyglyph(0xac, 0xffe2);	#fullwidth not
	#copyglyph(0xff5c, 0xffe4);	#fullwidth broken bar; not broken in 0xff5c

	copyglyph(0xffe5, 0xa5);	#yen

	copyglyph(0x30fb, 0x2022);	#bullet
	copyglyph(0x30fb, 0x2027);	#hyphenation point

	for ($i=1;$i<=0x5e;$i++) {
		copyglyph(0xff00+$i, 0x20+$i);	#fullwidth alnums to 0x20~
	}
}#JISX0208

if ($charset_registry =~ /^JISX0201\.1976/) {
	# hankaku-kana 0x01~0x1e not mapped by unicode
	# so manually copy them
	#copyglyph(0x01, 0x25c6);	#black diamond
	copyglyph(0x02, 0x2592);	#medium shade
	copyglyph(0x03, 0x2409);	#HT
	copyglyph(0x04, 0x240c);	#FF
	copyglyph(0x05, 0x240d);	#CR
	copyglyph(0x06, 0x240a);	#LF
	copyglyph(0x07, 0x02da);	#ring above
	#copyglyph(0x08, 0x00b1);	#+-
	copyglyph(0x09, 0x2424);	#NL
	copyglyph(0x0a, 0x2408);	#VT
	copyglyph(0x10, 0x23ba);	#scan line-1
	copyglyph(0x11, 0x23bb);	#scan line-3
	#copyglyph(0x12, );	#scan line-5
	copyglyph(0x13, 0x23bc);	#scan line-7
	copyglyph(0x14, 0x23bd);	#scan line-9
	copyglyph(0x1a, 0x2266);	#less-than over equal
	copyglyph(0x1b, 0x2267);	#greater-than over equal
	#copyglyph(0x1c, 0x03c0);	#pi
	copyglyph(0x1d, 0x2260);	#not equal
	#copyglyph(0x1f, 0x00a3);	#pound

	#copyglyph(0xff64, 0x3001);	#ideographic comma
	#copyglyph(0xff61, 0x3002);	#ideographic full stop
}#jisx0201


## next, open hanglg16.bdf and dump it, but
## //- add a scanline in BITMAP to make it 7x14
## - don't overwrite existing ./uchar/NNNNN
## - don't write out headers
if ($ARGV[0] eq "-b") {$bolden = 1; shift;}
if ($ARGV[0] eq "-br") {$bolden = 1; shift;}
if ($ARGV[0] eq "-bl") {$bolden = -1; shift;}
if ($ARGV[0] eq "+b") {$bolden = 0; shift;}
if ($ARGV[0] eq "-1") {$add_scanline = 1; shift;}
while ($bdf = shift(@ARGV)) {
print STDERR "Reading $bdf...\n";
open(BDF, "<", $bdf);
$state = "header";
#open(META, ">", "$charsdir/+HEADER") || die "cannot write +HEADER";
while(<BDF>) {
	#if (/^FONTBOUNDINGBOX (\d+) (\d+)/) {
	#	$font_bbw=$1; $font_bbh=$2; #used for sanity check
	#}
	if (/^COPYRIGHT "(.*)"/) {
		if ($copyright ne "") { $copyright .= "; "; }
		$copyright .= $1;
	}
	if (/^CHARSET_REGISTRY "(.*)"/) {
		$charset_registry = $1;
	}
	if (/^CHARSET_ENCODING "(.*)"/) {
		$charset_encoding = $1;
	}
	#if (/^DEFAULT_CHAR (\d+)/) {
	#	$default_char = $1;
	#}

	if (/^STARTCHAR/) {
		undef $code;
		$hcount=0;
		$cspec = $_;
		# read until ENDCHAR, and dump to separate file
		STARTCHAR: while(<BDF>) {
			if (/^ENCODING ([0-9]+)/) { $code = $1; }
			if (/^BBX ([0-9]+)/) {
				$bbw=$1;
				if ($bbw != $font_bbw) {
					print STDERR "WARNING: BBW $bbw != global $font_bbw in ENCODING $code\n";
				}
			}
			if (/^BITMAP/) {
				$cspec .= $_;
				while(<BDF>) {
					if (/^ENDCHAR/) {
						$cspec .= $_;
						last STARTCHAR;
					}
					$_ = embold($_, $bolden, $bbw);
					$cspec .= $_;
					$hcount++;
				}
			} #BITMAP
			last STARTCHAR if (/^ENDCHAR/);
			$cspec .= $_;
		} #STARTCHAR
		if ($add_scanline) {
			$cspec =~ s/^ENDCHAR/"00\n".$&/eim;
			$hcount++;
		}
		if (! defined($code)) { die "undefined ENCODING in STARTCHAR above line ".$.; }
		if ($hcount != $font_bbh) {
			print STDERR "WARNING: Height $hcount != global $font_bbh in ENCODING $code\n";
		}

		$ucode = charset2utf16c($code, "$charset_registry-$charset_encoding");
		if ($ucode != 0 || $code == 0 ) {
			$cspec =~ s/^ENCODING .*/"ENCODING $ucode"/mei;
			# dump to a separate file
			# don't overwrite
			if ( ! -e sprintf("$charsdir/%05s",$ucode)) {
				open(CSPEC,">", sprintf("$charsdir/%05s",$ucode));
				print CSPEC $cspec;
				close(CSPEC);
				#print STDERR "New glyph inserted: $ucode from charset_registry $charset_registry\n";
			}
		}

		# state transition
		if ($state eq "header") {
			#close(META);
			#open(META, ">", "$charsdir/=TRAILER") || die "cannot write =TRAILER";
			$state = "trailer";
		}

		next;
	}#STARTCHAR

	#print META $_;
} #<BDF>
} #wend shift filename.bdf

close BDF;

# re-count the chars
$charcount = 0;
opendir CHARS, "$charsdir" || die "no $charsdir ??";
while ($_ = readdir CHARS) { $charcount++ if /^[0-9]+/; }
#$charcount = scalar(grep {/^[0-9]+/} readdir CHARS); #slower list context
print STDERR "new CHARS: $charcount\n";
closedir CHARS;

## re-read the header, and modify
open(META,"<","$charsdir/+HEADER");
{
	local $/;	# full slurp mode
	$_ = <META>;
}
close META;
#print STDERR "HEADER: $_";
s/^FONT -(\w+)-([^-]+)-(\w+)-(R-Normal-\w*-\d+-\d+-\d+-\d+-\w+-\d+)-([^-]+)-(\d+)/"FONT -$1-$2-".(($bolden)?"Bold":$3)."-$4-ISO10646-1"/eim;
#s/^COMMENT [^\n]*\n//mg;
s/^DEFAULT_CHAR .*/"DEFAULT_CHAR $default_char"/eim;
s/^COPYRIGHT .*/"COPYRIGHT \"$copyright\""/eim;
#if (/^WEIGHT_NAME .*/m && $bolden) { s//WEIGHT_NAME "Bold"/m; }
#s/^FOUNDRY [^\n]*$/FOUNDRY "Misc"/im;
if ($bolden) { s/^WEIGHT_NAME .*/WEIGHT_NAME "Bold"/im;}
s/^ADD_STYLE_NAME .*/"ADD_STYLE_NAME \"$add_style\""/eim;
s/^CHARSET_REGISTRY [^\n]*/CHARSET_REGISTRY "ISO10646"/im;
s/^CHARSET_ENCODING .*/CHARSET_ENCODING "1"/im;
s/^CHARS [0-9]+/"CHARS $charcount"/eim;
open(META,">","$charsdir/+HEADER");
print META $_;
close META;

# output the result in +HEADER, NNNNN, =TRAILER order
opendir CHARS, "$charsdir" || die "no $charsdir ??";
foreach my $f (sort grep {!/^\./} readdir CHARS) {
	local $/;
	open F,"<","$charsdir/$f";
	print <F>;
	close F;
}
closedir CHARS;

# cleanup
system("rm -fr $charsdir");
